package mySynchMeans;

import java.util.concurrent.Semaphore;
import java.util.concurrent.atomic.AtomicInteger;


public class DormitorySemSem implements Dormitory {
    public boolean verbose; //to activate output msgs 
    public AtomicInteger occupiedSeats = new AtomicInteger(0); //state info: nr of occupied seats;
    public final Semaphore accessGate; //to control access to sleeping queue
    public final Semaphore sleepingPlaces; //to keep threads sleeping
    
   
    public DormitorySemSem(int nSeats) {
        this(nSeats, false);
    }
    
    public DormitorySemSem(int nSeats, boolean verbose) {
        this.verbose = verbose;
        sleepingPlaces = new Semaphore(0, true);
        accessGate = new Semaphore(nSeats,true);
    }
 
    
    @Override
    public void toSleep() throws InterruptedException{
        
        if(verbose)
            System.out.println("-->|    "+Thread.currentThread().getName()+
                    " asked to sleep   - # av. permits: "+
                    accessGate.availablePermits());
        
        accessGate.acquire(); //here, threads possibly wait for one available seat in the sleeping room
        
        if(verbose)
            System.out.println("   |--> "+Thread.currentThread().getName()+
                    " passed the gate  - # av. permits: "+
                    accessGate.availablePermits());
        
        occupiedSeats.incrementAndGet();
        
        sleepingPlaces.acquire(); //here, the thread blocks till the next wakeUpAll
               
        accessGate.release(); //upon leaving, a permit to enter is passed to accessGate
        
        if(verbose)
            System.out.println("        "+Thread.currentThread().getName()+
                    " has left ------> - # av. permits: "+
                    accessGate.availablePermits());
    }
    
  

    @Override
    public boolean tryToSleep() throws InterruptedException{
        
        if(verbose)
            System.out.println("~->|    "+Thread.currentThread().getName()+
                    " knocked to sleep - # av. permits: "+
                    accessGate.availablePermits());
        
        if(!accessGate.tryAcquire()) { //here, threads try to get an available seat in the sleeping room
            
            if(verbose)
                System.out.println(" <-|  ! "+Thread.currentThread().getName()+
                        " no seat availab. - # av. permits: "+
                        accessGate.availablePermits());    
            return false;
        }
        
        if(verbose)
            System.out.println("   |~-> "+Thread.currentThread().getName()+
                    " passed the gate  - # av. permits: "+
                    accessGate.availablePermits());
        
        occupiedSeats.incrementAndGet();
        
        sleepingPlaces.acquire(); //here, the thread blocks till the next wakeUpAll
               
        accessGate.release(); //upon leaving, a permit to enter is passed to accessGate
        
        if(verbose)
            System.out.println("        "+Thread.currentThread().getName()+
                    " has left ------> - # av. permits: "+
                    accessGate.availablePermits());
        
        return true;
    }


    
    @Override
    public int wakeUpAll() throws InterruptedException {
        
        int foundOccupied;

        foundOccupied = occupiedSeats.getAndSet(0);
        if (foundOccupied>0) {
            sleepingPlaces.release(foundOccupied);
        }

        if(verbose){
            System.out.println("\n--O-O-->> wakeUpAll: #released="+Integer.toString(foundOccupied)+"\n");
        }

        return foundOccupied;
    }

    @Override
    public void setVerbose(boolean verbose) {
        this.verbose = verbose;
    }
    
    

}
